//
// module data
// (c) 10.11.2007 Ingo D. Rullhusen
//

#include <complex>
#include <string>
#include <iostream>
#include <fstream>
#include <vector>
#include <algorithm>

#include <math.h>
#include <float.h>
#include <sys/stat.h>


#include "module_data.h"
#include "rotmat.h"
#include "triquad.h"


static const double DIST_EPSILON = 1e-10;

static double HYPOT( double x, double y, double z ){ return(sqrt(x*x+y*y+z*z)); }


ModuleData::ModuleData()
{
   off_file_name = "";
   mesh_unit = 1.0;

   cp_x = cp_y = cp_z = 0.0;
   theta = phi = psi = 0.0;

   grf = NULL;

   plist.clear();
   tlist.clear();
   velist.clear();

   LTbfktlist.clear();


   Iinclist.clear();


   J_vect_file_name = "";
   JPlist.clear();
   EPlist.clear();
   UPlist.clear();


   setup_exvec();

   Zrwg = NULL;
   IVrwg = NULL;
}

ModuleData::ModuleData( const ModuleData *d )
{
   off_file_name = d->off_file_name;
   mesh_unit = d->mesh_unit;

   cp_x = d->cp_x;
   cp_y = d->cp_y;
   cp_z = d->cp_z;
   theta = d->theta;
   phi = d->phi;
   psi = d->psi;

   if ( d->grf )
      grf = new GREEN( *d->grf );
   else
      grf = NULL;

   plist = d->plist;
   tlist = d->tlist;
   velist = d->velist;

   LTbfktlist = d->LTbfktlist;


   Iinclist = d->Iinclist;


   J_vect_file_name = d->J_vect_file_name;
   JPlist = d->JPlist;
   EPlist = d->EPlist;
   UPlist = d->UPlist;


   setup_exvec();


   if ( d->Zrwg )
      Zrwg = new SparseMatrix<SEP_Z_VAL>( d->Zrwg );
   else
      Zrwg = NULL;

   if ( d->IVrwg )
   {
      IVrwg = new SEP_Z_VAL[ velist.size() ];
      for ( unsigned int i=0; i<velist.size(); ++i ) IVrwg[i] = d->IVrwg[i];
   }
   else
      IVrwg = NULL;
}

ModuleData::~ModuleData()
{
   delete grf;
   grf = NULL;

   delete Zrwg;
   Zrwg = NULL;

   delete [] IVrwg;
   IVrwg = NULL;
}



class SORTpoint
{
   public:
      int no;
      double x, y, z;

      SORTpoint( int no_in=-1, double x_in=0.0, double y_in=0.0, double z_in=0.0 )
      : no(no_in), x(x_in), y(y_in), z(z_in)
      {}

      bool operator> ( const SORTpoint& sp2 ) const
      {
         if ( x > sp2.x )
            return(true);
         else if ( x == sp2.x )
            if ( y > sp2.y )
               return(true);
            else if ( y == sp2.y )
               if ( z > sp2.z )
                  return(true);
/**** doesn't work
         if ( x - sp2.x > DBL_EPSILON )
            return(true);
         else if ( y - sp2.y > DBL_EPSILON )   
            return(true);
         else if ( z > sp2.z )   
            return(true);
****/

         return(false);
      };

      bool operator< ( const SORTpoint& sp2 ) const
      {
         if ( x < sp2.x )
            return(true);
         else if ( x == sp2.x )
            if ( y < sp2.y )   
               return(true);   
            else if ( y == sp2.y )
               if ( z < sp2.z )
                  return(true);
/**** doesn't work
         if ( x - sp2.x < -DBL_EPSILON )
            return(true);
         else if ( y - sp2.y < -DBL_EPSILON )
            return(true);
         else if ( z < sp2.z )
            return(true);
****/

         return(false);
      };
};

class SORTdata
{
   public:
      int p1, p2;
      int p3, t;

      SORTdata( int p1_in=-1, int p2_in=-1, int p3_in=-1, int t_in=-1 )
      : p1(p1_in), p2(p2_in), p3(p3_in), t(t_in)
      {}

      bool operator> ( const SORTdata& sd2 ) const
      {
         if ( p1 > sd2.p1 )
            return(true);
         else if ( p1 == sd2.p1 )
            if ( p2 > sd2.p2 )
               return(true);
         return(false);
      };

      bool operator< ( const SORTdata& sd2 ) const
      {
         if ( p1 < sd2.p1 )
            return(true);
         else if ( p1 == sd2.p1 )
            if ( p2 < sd2.p2 )
               return(true);
         return(false);
      };
};

void ModuleData::read_off_file( string &message )
{
   char cbuff;
   int pno_t, tno_t, eno_t, n;
   double x, y, z, rmat[3][3];
   string strbuff;
   ifstream off_file;
   SORTpoint sp_t;
   TRIANGLE t_t;
   vector<SORTpoint> splist;

   message = "";

   splist.clear();

   plist.clear();
   tlist.clear();
   velist.clear();

   rotmat( rmat, mesh_unit, psi, 0.0, 0.0, theta, phi );

   off_file.open( off_file_name.c_str(), ios::in );
   if ( !off_file.is_open() )
   {
      throw( "cannot open OFF file" );
   }

// head
   off_file >> strbuff;

   if ( strbuff != "OFF" )
   {
      throw( "not an OFF file" );
   }
   off_file >> strbuff;
   pno_t = atoi(strbuff.c_str());
   off_file >> strbuff;
   tno_t = atoi(strbuff.c_str());
   off_file >> strbuff;
   eno_t = atoi(strbuff.c_str());

// points
   for ( int p=0; p<pno_t; p++ )
   {
      off_file >> strbuff;
      x = atof(strbuff.c_str());
      off_file >> strbuff;
      y = atof(strbuff.c_str());
      off_file >> strbuff;
      z = atof(strbuff.c_str());

      sp_t.no = p;
      sp_t.x = rmat[0][0]*x + rmat[0][1]*y + rmat[0][2]*z + cp_x;
      sp_t.y = rmat[1][0]*x + rmat[1][1]*y + rmat[1][2]*z + cp_y;
      sp_t.z = rmat[2][0]*x + rmat[2][1]*y + rmat[2][2]*z + cp_z;

      splist.push_back( sp_t );
   }

// triangles
   for ( int t=0; t<tno_t; t++ )
   {
      off_file >> strbuff;
      n = atoi(strbuff.c_str());
      if ( n == 3 )
      {
         off_file >> strbuff;
         t_t.p1 = atoi(strbuff.c_str());
         off_file >> strbuff;
         t_t.p2 = atoi(strbuff.c_str());
         off_file >> strbuff;
         t_t.p3 = atoi(strbuff.c_str());

         t_t.ve1 = -1;
         t_t.ve2 = -1;
         t_t.ve3 = -1;
         t_t.sign.sp1 = 0;
         t_t.sign.sp2 = 0;
         t_t.sign.sp3 = 0;

         if ( t_t.p1 < 0 ||
              t_t.p2 < 0 ||
              t_t.p3 < 0 ||
              (unsigned)abs(t_t.p1) >= splist.size() ||
              (unsigned)abs(t_t.p2) >= splist.size() ||
              (unsigned)abs(t_t.p3) >= splist.size()
            )
         {
            throw( "wrong point index" );
         }

         tlist.push_back( t_t );
      }
      else
      {
         throw( "non-triangle found" );
      }

      off_file.get(cbuff);
      while ( cbuff != '\n' && !off_file.eof() )
      {
         off_file.get(cbuff); 
      }

   }

   if ( off_file.fail() )
   {
      throw( "trouble with OFF file" );
   }

   off_file.close();

   if ( tlist.size() < 1 )
   {
      throw( "no triangles found" );
   }


//
// remove multiple points
//
   sort( splist.begin(), splist.end() );

   plist.clear();

   POINT p_t;
   int *permute, p;
   permute = new int[ splist.size() ];

   vector<SORTpoint>::iterator isp1, isp2;
   isp1 = splist.begin();

   p = 0;
   p_t.x = isp1->x;
   p_t.y = isp1->y;
   p_t.z = isp1->z;
   plist.push_back( p_t );
   permute[isp1->no] = p;

   for ( isp2 = isp1 + 1;
         isp2 != splist.end();
         isp1++, isp2++ )
   {
      if ( HYPOT( isp1->x-isp2->x, isp1->y-isp2->y, isp1->z-isp2->z ) > DIST_EPSILON )
      {
         p++;
         p_t.x = isp2->x;
         p_t.y = isp2->y;
         p_t.z = isp2->z;
         plist.push_back( p_t );
         permute[isp2->no] = p;
      }
      else
      {
         permute[isp2->no] = p;
      }
   }

   for ( TRIANGLE_LIST::iterator it = tlist.begin();
         it != tlist.end();
         it++ )
   {
      it->p1 = permute[it->p1];
      it->p2 = permute[it->p2];
      it->p3 = permute[it->p3];
   }

   delete [] permute;
   splist.clear();


//
// searching edges
//
   unsigned int t, veno;
   vector<SORTdata> SDlist;
   vEDGE vE_t;
   vector<SORTdata>::iterator sd1, sd2;

   veno = 0;

   SDlist.clear();

   for ( TRIANGLE_LIST::iterator it = tlist.begin();
         it != tlist.end();
         it++ )
   {
      t = it - tlist.begin();

// edge 1
      if ( it->p1 < it->p2 )
      {
         SDlist.push_back( SORTdata( it->p1, it->p2, it->p3, t ) );
      }
      else
      {
         SDlist.push_back( SORTdata( it->p2, it->p1, it->p3, t ) );
      }

// edge 2
      if ( it->p2 < it->p3 )
      {
         SDlist.push_back( SORTdata( it->p2, it->p3, it->p1, t ) );
      }
      else
      {
         SDlist.push_back( SORTdata( it->p3, it->p2, it->p1, t ) );
      }

// edge 3
      if ( it->p3 < it->p1 )
      {
         SDlist.push_back( SORTdata( it->p3, it->p1, it->p2, t ) );
      }
      else
      {
         SDlist.push_back( SORTdata( it->p1, it->p3, it->p2, t ) );
      }

   }


   sort( SDlist.begin(), SDlist.end() );


   sd1=SDlist.begin();
   while( sd1->p1<0 || sd1->p2<0 || sd1->p3<0 || sd1->t<0 ) sd1++;
   for ( sd2 = sd1 + 1;
         sd1 != SDlist.end() && sd2 != SDlist.end();
         sd1++, sd2++ )
   {
      if ( sd1->p1 == sd2->p1 && sd1->p2 == sd2->p2 )
      {
         vE_t.p1 = sd1->p1;
         vE_t.p2 = sd1->p2;
         vE_t.spm = sd1->p3;
         vE_t.spp = sd2->p3;
         vE_t.trim = sd1->t;
         vE_t.trip = sd2->t;
         velist.push_back( vE_t );

         if ( tlist[sd1->t].p1 == sd1->p3 )
         {
            tlist[sd1->t].ve1  = veno;
            tlist[sd1->t].sign.sp1 = -1;
         }
         else if ( tlist[sd1->t].p2 == sd1->p3 )
         {
            tlist[sd1->t].ve2  = veno;
            tlist[sd1->t].sign.sp2 = -1;
         }
         else if ( tlist[sd1->t].p3 == sd1->p3 )
         {
            tlist[sd1->t].ve3  = veno;
            tlist[sd1->t].sign.sp3 = -1;
         }
         else
         {
            cerr << "Error in ModuleData::ModuleData(): internal line "
                 << __LINE__ << "!" << endl;
            exit(1);
         }

         if ( tlist[sd2->t].p1 == sd2->p3 )
         {
            tlist[sd2->t].ve1  = veno;
            tlist[sd2->t].sign.sp1 = +1;
         }
         else if ( tlist[sd2->t].p2 == sd2->p3 )
         {
            tlist[sd2->t].ve2  = veno;
            tlist[sd2->t].sign.sp2 = +1;
         }
         else if ( tlist[sd2->t].p3 == sd2->p3 )
         {
            tlist[sd2->t].ve3  = veno;
            tlist[sd2->t].sign.sp3 = +1;
         }
         else
         {
            cerr << "Error in ModuleData::ModuleData(): internal line "
                 << __LINE__ << "!" << endl;
            exit(1);
         }

         ++sd1; ++sd2; ++veno;
      }
   }

   SDlist.clear();


   LTbfktlist.clear();
//
// looking for loops
//
   int startt, nxtt;
   LTbfkt LTb_t;

   for ( vEDGE_LIST::iterator it = velist.begin();
         it != velist.end();
         ++it )
   {
      t = it - velist.begin();

// point 1
      SDlist.push_back( SORTdata( it->p1, it->p2, 0, t ) );

// point 2
      SDlist.push_back( SORTdata( it->p2, it->p1, 0, t ) );
   }

   sort( SDlist.begin(), SDlist.end() );

   for ( sd1=SDlist.begin(); sd1 != SDlist.end(); sd1=sd2 )
   {
      for ( sd2=sd1+1; sd2!=SDlist.end(); ++sd2 )
      {
         if ( sd1->p1 != sd2->p1 ) break;
      }

      LTb_t.count = 0;
      LTb_t.norm.clear();
      LTb_t.bfkt_in_LT.clear();

// first
      startt = velist[ sd1->t ].trip;

      LTb_t.bfkt_in_LT.push_back( sd1->t );
      LTb_t.norm.push_back( 1.0 );
      ++LTb_t.count;

      nxtt = velist[ sd1->t ].trim;

      sd1->t = -1;

// next
      vector<SORTdata>::iterator sdn = sd1;
      while ( sdn != sd2 )
      {
         if ( sdn->t < 0 )
         {
            ++sdn;
            continue;
         }

         if      ( velist[sdn->t].trip == nxtt )
         {
            LTb_t.bfkt_in_LT.push_back( sdn->t );
            LTb_t.norm.push_back( 1.0 );
            ++LTb_t.count;

            nxtt = velist[ sdn->t ].trim;

            sdn->t = -1;
            sdn = sd1;
         }
         else if ( velist[sdn->t].trim == nxtt )
         {
            LTb_t.bfkt_in_LT.push_back( sdn->t );
            LTb_t.norm.push_back( -1.0 );
            ++LTb_t.count;

            nxtt = velist[ sdn->t ].trip;

            sdn->t = -1;
            sdn = sd1;
         }

         ++sdn;
      }

// closed?
      if ( nxtt == startt ) LTbfktlist.push_back( LTb_t );
   }

//
// looking for tree basis
//
   int sedge, k, kn;
   bool *uT = new bool[tlist.size()];
   for ( unsigned int i=0; i<tlist.size(); i++ ) uT[i] = true;

   k = kn = LTbfktlist.size();

   sedge = 0;

   while ( sedge > -1 )
   {

      LTb_t.count = 1;
      LTb_t.norm.clear();
      LTb_t.bfkt_in_LT.clear();

      LTb_t.bfkt_in_LT.push_back( sedge );
      LTb_t.norm.push_back( 1.0 );

      LTbfktlist.push_back( LTb_t );
      ++k;

      uT[ velist[sedge].trip ] = false;
      uT[ velist[sedge].trim ] = false;

      t = velist[sedge].trim;
      sedge = -1;
      if (  tlist[ t ].sign.sp1 != 0 )
      {
         if ( uT[ velist[ tlist[ t ].ve1 ].trip ] ||
              uT[ velist[ tlist[ t ].ve1 ].trim ]   ) sedge = tlist[ t ].ve1;
      }
      if (  tlist[ t ].sign.sp2 != 0 )
      {
         if ( uT[ velist[ tlist[ t ].ve2 ].trip ] || 
              uT[ velist[ tlist[ t ].ve2 ].trim ]   ) sedge = tlist[ t ].ve2;
      }
      if (  tlist[ t ].sign.sp3 != 0 )
      {
         if ( uT[ velist[ tlist[ t ].ve3 ].trip ] ||
              uT[ velist[ tlist[ t ].ve3 ].trim ]   ) sedge = tlist[ t ].ve3;
      }

// branch start
      if ( sedge < 0 )
      {
         for ( int kk=k-1; kk>kn-1; --kk )
         {

            t = velist[ LTbfktlist[kk].bfkt_in_LT[0] ].trim;
            if (  tlist[ t ].sign.sp1 != 0 )
            {
               if ( uT[ velist[ tlist[ t ].ve1 ].trip ] ||
                    uT[ velist[ tlist[ t ].ve1 ].trim ]   ) sedge = tlist[ t ].ve1;
            }
            if (  tlist[ t ].sign.sp2 != 0 )
            {
               if ( uT[ velist[ tlist[ t ].ve2 ].trip ] || 
                    uT[ velist[ tlist[ t ].ve2 ].trim ]   ) sedge = tlist[ t ].ve2;
            }
            if (  tlist[ t ].sign.sp3 != 0 )
            {
               if ( uT[ velist[ tlist[ t ].ve3 ].trip ] ||
                    uT[ velist[ tlist[ t ].ve3 ].trim ]   ) sedge = tlist[ t ].ve3;
            }

            t = velist[ LTbfktlist[kk].bfkt_in_LT[0] ].trip;
            if (  tlist[ t ].sign.sp1 != 0 )
            {
               if ( uT[ velist[ tlist[ t ].ve1 ].trip ] ||
                    uT[ velist[ tlist[ t ].ve1 ].trim ]   ) sedge = tlist[ t ].ve1;
            }
            if (  tlist[ t ].sign.sp2 != 0 )
            {
               if ( uT[ velist[ tlist[ t ].ve2 ].trip ] || 
                    uT[ velist[ tlist[ t ].ve2 ].trim ]   ) sedge = tlist[ t ].ve2;
            }
            if (  tlist[ t ].sign.sp3 != 0 )
            {
               if ( uT[ velist[ tlist[ t ].ve3 ].trip ] ||
                    uT[ velist[ tlist[ t ].ve3 ].trim ]   ) sedge = tlist[ t ].ve3;
            }

         }
      }

// patch start
      if ( sedge < 0 )
      {
         for ( t=0; (t<tlist.size()) && !uT[t] ; ++t ){};

         if ( t < tlist.size() )
         {
            if ( tlist[ t ].sign.sp1 != 0 ) sedge = tlist[ t ].ve1;
            if ( tlist[ t ].sign.sp2 != 0 ) sedge = tlist[ t ].ve2;
            if ( tlist[ t ].sign.sp3 != 0 ) sedge = tlist[ t ].ve3;
         }
      }

   }


   SDlist.clear();

   delete [] uT;

   if ( LTbfktlist.size() != veno )
   {
      if ( LTbfktlist.size() > veno )  // closed surfaces (?)
      {
// brush under the carpet and hope only
         if ( (LTbfktlist.size() - veno) != 1 )
            message += "topological problem brushed under the carpet; ";

         k = kn / ( LTbfktlist.size() - veno );
         for ( int l=kn-1; l>0; l -= k )
         {
            while ( LTbfktlist[l].count < 2 && l > 0 ) --l;  // really loop?
            LTbfktlist.erase( LTbfktlist.begin()+l );  // remove one loop per body
         }
      }
      else throw("LT error");
   }

/*****
#warning RWG als pseudo tree
message += "plain RWG as pseudo LT; ";
LTbfktlist.resize(veno);
for ( unsigned int i=0; i<veno; i++ )
{
LTbfktlist[i].count = 1;
LTbfktlist[i].norm.clear();
LTbfktlist[i].norm.push_back(1.0);
LTbfktlist[i].bfkt_in_LT.clear();
LTbfktlist[i].bfkt_in_LT.push_back(i);
}
*****/


}


#ifdef HAVE_LIBGL
void ModuleData::ogl_triangles( GLuint list )
{
   const GLfloat clr_1[4] = { 0.12, 0.90, 0.90, 1.0 };
   const GLfloat clr_2[4] = { 0.78, 0.20, 0.12, 1.0 };

   double u[3], v[3], w[3], a;

   glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, clr_1 );
   glMaterialfv( GL_BACK, GL_AMBIENT_AND_DIFFUSE, clr_2 );

   glBegin(GL_TRIANGLES);

   for ( vector<TRIANGLE>::iterator it = tlist.begin();
         it != tlist.end();  
         it++ )
   {

      u[0] = plist[ it->p2 ].x -  plist[ it->p1 ].x;
      u[1] = plist[ it->p2 ].y -  plist[ it->p1 ].y;
      u[2] = plist[ it->p2 ].z -  plist[ it->p1 ].z;

      v[0] = plist[ it->p3 ].x -  plist[ it->p2 ].x;
      v[1] = plist[ it->p3 ].y -  plist[ it->p2 ].y;
      v[2] = plist[ it->p3 ].z -  plist[ it->p2 ].z;

      w[0] = u[1]*v[2] - u[2]*v[1];
      w[1] = u[2]*v[0] - u[0]*v[2];
      w[2] = u[0]*v[1] - u[1]*v[0];

      a = sqrt( w[0]*w[0] + w[1]*w[1] + w[2]*w[2] );

      glNormal3f( w[0]/a, w[1]/a, w[2]/a );


      glVertex3f( plist[it->p1].x,  plist[it->p1].y, plist[it->p1].z );
      glVertex3f( plist[it->p2].x,  plist[it->p2].y, plist[it->p2].z );
      glVertex3f( plist[it->p3].x,  plist[it->p3].y, plist[it->p3].z );
   }

   glEnd();
}
#endif


void ModuleData::triangle_center( int tno, POINT &cp ) const
{
#ifdef BOUNDARY_CHECK
   if ( tno < 0 || tno >= (int)tlist.size() )
   {
      cerr << "Error in ModuleData::triangle_center(): tno out of bounds!"
           << endl;
      exit(1);
   }
#endif
   cp.x = ( plist[ tlist[tno].p1 ].x
            + plist[ tlist[tno].p2 ].x
              + plist[ tlist[tno].p3 ].x ) / 3.0;
   cp.y = ( plist[ tlist[tno].p1 ].y
            + plist[ tlist[tno].p2 ].y
              + plist[ tlist[tno].p3 ].y ) / 3.0;
   cp.z = ( plist[ tlist[tno].p1 ].z
            + plist[ tlist[tno].p2 ].z
              + plist[ tlist[tno].p3 ].z ) / 3.0;
}

bool ModuleData::triangle_neigh( int tno1, int tno2 ) const
{
   POINT cp1, cp2;
   double r1, r2, R;

   triangle_center( tno1, cp1 );
   triangle_center( tno2, cp2 );

   r1 = HYPOT( cp1.x-plist[ tlist[tno1].p1 ].x,
               cp1.y-plist[ tlist[tno1].p1 ].y,
               cp1.z-plist[ tlist[tno1].p1 ].z  );
   R = HYPOT( cp1.x-plist[ tlist[tno1].p2 ].x,
              cp1.y-plist[ tlist[tno1].p2 ].y,
              cp1.z-plist[ tlist[tno1].p2 ].z  );
   r1 = ( R > r1 ? R : r1 );
   R = HYPOT( cp1.x-plist[ tlist[tno1].p3 ].x,
              cp1.y-plist[ tlist[tno1].p3 ].y,
              cp1.z-plist[ tlist[tno1].p3 ].z  );
   r1 = ( R > r1 ? R : r1 );

   r2 = HYPOT( cp2.x-plist[ tlist[tno2].p1 ].x,
               cp2.y-plist[ tlist[tno2].p1 ].y,
               cp2.z-plist[ tlist[tno2].p1 ].z  );
   R = HYPOT( cp2.x-plist[ tlist[tno2].p2 ].x,
              cp2.y-plist[ tlist[tno2].p2 ].y,
              cp2.z-plist[ tlist[tno2].p2 ].z  );
   r2 = ( R > r2 ? R : r2 );
   R = HYPOT( cp2.x-plist[ tlist[tno2].p3 ].x,
              cp2.y-plist[ tlist[tno2].p3 ].y,
              cp2.z-plist[ tlist[tno2].p3 ].z  ); 
   r2 = ( R > r2 ? R : r2 );

   if ( HYPOT( cp1.x-cp2.x, cp1.y-cp2.y, cp1.z-cp2.z ) > r1+r2 ) return(false);

   return(true);
}


double ModuleData::triangle_area( int tno ) const
{
   double nx, ny, nz;
   return( triangle_normal( tno, nx, ny, nz ) );
}


int ModuleData::near_triangle( POINT &p ) const
{
   int nt = 0;
   POINT tp;

   triangle_center( nt, tp );

   double dt, d = HYPOT( p.x-tp.x, p.y-tp.y, p.z-tp.z );

   for ( int t=1; t<(int)tlist.size(); t++ )
   {
      triangle_center( t, tp );
      if ( d > ( dt = HYPOT( p.x-tp.x, p.y-tp.y, p.z-tp.z ) ) )
      {
         d = dt;
         nt = t;
      }
   }

   return(nt);
}


void ModuleData::setup_exvec()
{
   double N, n[3], r[3];

   for ( Jprobe_LIST::iterator jp = JPlist.begin();
         jp != JPlist.end() && tlist.size() > 0;
         ++jp )
   {
      jp->tri = near_triangle( jp->p );

      triangle_normal( jp->tri, n[0], n[1], n[2] );

      r[0] = jp->p.x - plist[ tlist[jp->tri].p1 ].x;
      r[1] = jp->p.y - plist[ tlist[jp->tri].p1 ].y;
      r[2] = jp->p.z - plist[ tlist[jp->tri].p1 ].z;

      N = r[0]*n[0] + r[1]*n[1] + r[2]*n[2];

      jp->p.x -= N*n[0];
      jp->p.y -= N*n[1];
      jp->p.z -= N*n[2];
   }

}


double ModuleData::triangle_normal( int tno, double &nx, double &ny, double &nz ) const
{
   double a[3], b[3], AR2;

#ifdef BOUNDARY_CHECK
   if ( tno < 0 || tno >= (int)tlist.size() )
   {
      cerr << "Error in ModuleData::triangle_normal(): tno out of bounds!"
           << endl;
      exit(1);
   }
#endif

   a[0] = plist[ tlist[tno].p2 ].x - plist[ tlist[tno].p1 ].x;
   a[1] = plist[ tlist[tno].p2 ].y - plist[ tlist[tno].p1 ].y;
   a[2] = plist[ tlist[tno].p2 ].z - plist[ tlist[tno].p1 ].z;

   b[0] = plist[ tlist[tno].p3 ].x - plist[ tlist[tno].p1 ].x;
   b[1] = plist[ tlist[tno].p3 ].y - plist[ tlist[tno].p1 ].y;
   b[2] = plist[ tlist[tno].p3 ].z - plist[ tlist[tno].p1 ].z;

   nx = a[1]*b[2] - a[2]*b[1];
   ny = a[2]*b[0] - a[0]*b[2];
   nz = a[0]*b[1] - a[1]*b[0];

   AR2 = HYPOT( nx, ny, nz );
   nx /= AR2;
   ny /= AR2;
   nz /= AR2;

   return( AR2 * 0.5 );
}


void ModuleData::update_Iinc()
{
   double dv, a[3], b[3], sangle;

   if ( plist.size() < 1 || tlist.size() < 1 ) return;  // geo not ready

   for ( Iinc_LIST::iterator ici = Iinclist.begin();
         ici != Iinclist.end();
         ++ici )
   {

// find node
      dv = DBL_MAX;
      ici->sp = 0;
      for ( POINT_LIST::iterator pi = plist.begin();
            pi != plist.end();
            ++pi )
      {
         if ( HYPOT( pi->x - ici->x, pi->y - ici->y, pi->z - ici->z )
              < dv )
         {
            dv = HYPOT( pi->x - ici->x, pi->y - ici->y, pi->z - ici->z );
            ici->sp = pi - plist.begin();
         }
      }

// find edges
      ici->tp1.clear();
      ici->tp2.clear();
      ici->trino.clear();
      ici->angle.clear();
      ici->area.clear();

      for ( TRIANGLE_LIST::iterator ti = tlist.begin();
            ti != tlist.end();
            ++ti )
      {
         if ( ti->p1 == ici->sp )
         {
            ici->tp1.push_back( ti->p2 );
            ici->tp2.push_back( ti->p3 );
            ici->trino.push_back( ti - tlist.begin() );
         }
         else if ( ti->p2 == ici->sp )
         {
            ici->tp1.push_back( ti->p1 );
            ici->tp2.push_back( ti->p3 );
            ici->trino.push_back( ti - tlist.begin() );
         }
         else if ( ti->p3 == ici->sp )
         {
            ici->tp1.push_back( ti->p1 );
            ici->tp2.push_back( ti->p2 );
            ici->trino.push_back( ti - tlist.begin() );
         }
         else continue;

         a[0] = plist[ ici->tp1.back() ].x - plist[ ici->sp ].x;
         a[1] = plist[ ici->tp1.back() ].y - plist[ ici->sp ].y;
         a[2] = plist[ ici->tp1.back() ].z - plist[ ici->sp ].z;
         dv = HYPOT( a[0], a[1], a[2] );
         a[0] /= dv;
         a[1] /= dv;
         a[2] /= dv;

         b[0] = plist[ ici->tp2.back() ].x - plist[ ici->sp ].x;
         b[1] = plist[ ici->tp2.back() ].y - plist[ ici->sp ].y;
         b[2] = plist[ ici->tp2.back() ].z - plist[ ici->sp ].z;
         dv = HYPOT( b[0], b[1], b[2] );
         b[0] /= dv;
         b[1] /= dv;
         b[2] /= dv;

         ici->angle.push_back(
            fabs(acos( a[0]*b[0] + a[1]*b[1] + a[2]*b[2] ))
                             );

         ici->area.push_back( triangle_area( ici->trino.back() ) );
      }

      ici->count = ici->tp1.size();
#ifdef BOUNDARY_CHECK   
      if ( ici->count != (int)ici->tp2.size()   ||
           ici->count != (int)ici->trino.size() ||
           ici->count != (int)ici->angle.size() ||
           ici->count != (int)ici->area.size()     )
      {
         cerr << "Error in ModuleData::update_Iinc(): count inconsistent!"
              << endl;
         exit(1);
      }
#endif

      sangle = 0.0;
      for ( int i=0; i<ici->count; ++i )
      {
         sangle += ici->angle[i];
      }

      ici->Jampl_fac = - ici->Iampl / ( 2.0*sangle );
   }

}


void ModuleData::bfkt_E( complex<double> E[3], int j, POINT ap ) const
{
   triquad iI(7);
   int qtri;
   double rv2[3];
   complex<double> Gd[3][3];
   POINT qp;

   E[0] = E[1] = E[2] = complex<double>(0.0,0.0);

#ifdef BOUNDARY_CHECK
   if ( j < 0 || j >= (int)velist.size() )
   {
      cerr << "Error in ModuleData::bfkt_E(): index out of bounds!"
           << endl;
      exit(1);
   }
#endif

// triangle +
   qtri = velist[j].trip;

   for ( int ip2=0; ip2<iI.no; ip2++ )
   {
      qp.x = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].x
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].x
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].x;
      qp.y = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].y   
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].y 
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].y;
      qp.z = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].z   
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].z 
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].z;


      rv2[0] = ( qp.x - plist[ velist[j].spp ].x ) * 0.5;
      rv2[1] = ( qp.y - plist[ velist[j].spp ].y ) * 0.5;
      rv2[2] = ( qp.z - plist[ velist[j].spp ].z ) * 0.5;

      grf->EF_green( Gd, ap, qp );

      E[0] += iI.weights[ip2] *
              ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /triangle
      E[1] += iI.weights[ip2] *
              ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /triangle
      E[2] += iI.weights[ip2] *
              ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /triangle
   }

// triangle -
   qtri = velist[j].trim;

   for ( int ip2=0; ip2<iI.no; ip2++ )
   {
      qp.x = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].x
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].x
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].x;
      qp.y = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].y   
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].y 
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].y;
      qp.z = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].z   
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].z 
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].z;


      rv2[0] = ( qp.x - plist[ velist[j].spm ].x ) * 0.5;
      rv2[1] = ( qp.y - plist[ velist[j].spm ].y ) * 0.5;
      rv2[2] = ( qp.z - plist[ velist[j].spm ].z ) * 0.5;

      grf->EF_green( Gd, ap, qp );

      E[0] -= iI.weights[ip2] *
              ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /triangle
      E[1] -= iI.weights[ip2] *
              ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /triangle
      E[2] -= iI.weights[ip2] *
              ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /triangle
   }

}



void ModuleData::inc_E( complex<double> E[3], int ni, POINT ap ) const
{
   triquad iI(7);
   int qtri;
   double rv2[3];
   complex<double> Gd[3][3];
   POINT qp;

#ifdef BOUNDARY_CHECK
   if ( ni < 0 || ni >= (int)Iinclist.size() )
   {   
      cerr << "Error in ModuleData::inc_E(): index out of bounds!"          
           << endl;   
      exit(1);   
   }   
#endif 

   Iinc *ici = const_cast<Iinc*>(&Iinclist[ni]);

   E[0] = E[1] = E[2] = complex<double>(0.0,0.0);

   for ( int t=0; t<ici->count; ++t )
   {
   qtri = ici->trino[t];

   for ( int ip2=0; ip2<iI.no; ip2++ )
   {
      qp.x = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].x
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].x
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].x;
      qp.y = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].y   
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].y 
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].y;
      qp.z = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].z   
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].z 
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].z;


      grf->EF_green( Gd, ap, qp );


// edge 1
      rv2[0] = ( qp.x - plist[ ici->tp1[t] ].x ) * 0.5;
      rv2[1] = ( qp.y - plist[ ici->tp1[t] ].y ) * 0.5;
      rv2[2] = ( qp.z - plist[ ici->tp1[t] ].z ) * 0.5;

      E[0] += iI.weights[ip2] * ici->angle[t] *
              ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /triangle
      E[1] += iI.weights[ip2] * ici->angle[t] *
              ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /triangle
      E[2] += iI.weights[ip2] * ici->angle[t] *
              ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /triangle

// edge 2
      rv2[0] = ( qp.x - plist[ ici->tp2[t] ].x ) * 0.5;
      rv2[1] = ( qp.y - plist[ ici->tp2[t] ].y ) * 0.5;
      rv2[2] = ( qp.z - plist[ ici->tp2[t] ].z ) * 0.5;

      E[0] += iI.weights[ip2] * ici->angle[t] *
              ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /triangle
      E[1] += iI.weights[ip2] * ici->angle[t] *
              ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /triangle
      E[2] += iI.weights[ip2] * ici->angle[t] *
              ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /triangle
   }

   }

}


void ModuleData::bfkt_pot( complex<double> A[3], complex<double> &phi,
                           double sgr, int e2, POINT ap )
{
   const triquad iI(7);
   const complex<double> cNULL(0.0,0.0);
   const complex<double> cONE(1.0,0.0);

   bool singu;
   int qtri;
   double rv2[3], area;
   complex<double> Gd[3][3], Gs, v[3], J[3], div, J0[3], div0;
   POINT qp;

   A[0] = cNULL;
   A[1] = cNULL;
   A[2] = cNULL;
   phi = cNULL;

//
// bfkt
   if ( e2 > -1 && e2 < (int)velist.size() )
   {

// triangle 2 p
   qtri = velist[e2].trip;
   triangle_center( qtri, qp );

   if ( HYPOT( ap.x-qp.x, ap.y-qp.y, ap.z-qp.z ) < sgr )
   {
      singu = true;
      J0[0] = ( qp.x - plist[ velist[e2].spp ].x ) * 0.5;
      J0[1] = ( qp.y - plist[ velist[e2].spp ].y ) * 0.5;
      J0[2] = ( qp.z - plist[ velist[e2].spp ].z ) * 0.5;
      div0 = cONE;
   }
   else
      singu = false;

   for ( int ip2=0; ip2<iI.no; ip2++ )
   {
      qp.x = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].x
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].x
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].x;
      qp.y = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].y
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].y
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].y;
      qp.z = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].z
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].z
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].z;


      rv2[0] = ( qp.x - plist[ velist[e2].spp ].x ) * 0.5;
      rv2[1] = ( qp.y - plist[ velist[e2].spp ].y ) * 0.5;
      rv2[2] = ( qp.z - plist[ velist[e2].spp ].z ) * 0.5;

      div = cONE;

      if ( singu )
      {
         J[0] = rv2[0];
         J[1] = rv2[1];
         J[2] = rv2[2];
         grf->EP_green_ex( v, Gs, J, div, J0, div0, ap, qp );

         A[0] += iI.weights[ip2] * v[0];  // /triangle
         A[1] += iI.weights[ip2] * v[1];  // /triangle
         A[2] += iI.weights[ip2] * v[2];  // /triangle

         phi += iI.weights[ip2] * Gs;  // /triangle
      }
      else
      {
         grf->EP_green( Gd, Gs, ap, qp );

         A[0] += iI.weights[ip2] *
                 ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /triangle
         A[1] += iI.weights[ip2] *
                 ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /triangle
         A[2] += iI.weights[ip2] *
                 ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /triangle

         phi += iI.weights[ip2] * Gs * div;  // /triangle
      }
   }

   if ( singu )
   {
      area = triangle_area(qtri);
      J0[0] /= area;
      J0[1] /= area;
      J0[2] /= area;
      div0  /= area;

      grf->EP_green_re( v, div, J0, div0,
                              plist[ tlist[qtri].p1 ],
                                 plist[ tlist[qtri].p2 ],
                                    plist[ tlist[qtri].p3 ], ap );
      A[0] += v[0];
      A[1] += v[1];
      A[2] += v[2];
      phi += div;
   }

// triangle 2 m
   qtri = velist[e2].trim;
   triangle_center( qtri, qp );

   if ( HYPOT( ap.x-qp.x, ap.y-qp.y, ap.z-qp.z ) < sgr )
   {
      singu = true;
      J0[0] = ( plist[ velist[e2].spm ].x - qp.x ) * 0.5;
      J0[1] = ( plist[ velist[e2].spm ].y - qp.y ) * 0.5;
      J0[2] = ( plist[ velist[e2].spm ].z - qp.z ) * 0.5;
      div0 = -cONE;
   }
   else
      singu = false;

   for ( int ip2=0; ip2<iI.no; ip2++ )
   {
      qp.x = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].x
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].x
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].x;
      qp.y = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].y
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].y
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].y;
      qp.z = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].z
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].z
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].z;

      rv2[0] = ( plist[ velist[e2].spm ].x - qp.x ) * 0.5;
      rv2[1] = ( plist[ velist[e2].spm ].y - qp.y ) * 0.5;
      rv2[2] = ( plist[ velist[e2].spm ].z - qp.z ) * 0.5;

      div = -cONE;

      if ( singu )
      {
         J[0] = rv2[0];
         J[1] = rv2[1];
         J[2] = rv2[2];
         grf->EP_green_ex( v, Gs, J, div, J0, div0, ap, qp );

         A[0] += iI.weights[ip2] * v[0];  // /triangle
         A[1] += iI.weights[ip2] * v[1];  // /triangle
         A[2] += iI.weights[ip2] * v[2];  // /triangle

         phi += iI.weights[ip2] * Gs;  // /triangle
      }
      else
      {
         grf->EP_green( Gd, Gs, ap, qp );

         A[0] += iI.weights[ip2] *
                 ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /triangle
         A[1] += iI.weights[ip2] *
                 ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /triangle
         A[2] += iI.weights[ip2] *
                 ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /triangle

         phi += iI.weights[ip2] * Gs * div;  // /triangle
      }
   }

   if ( singu )
   {
      area = triangle_area(qtri);
      J0[0] /= area;
      J0[1] /= area;
      J0[2] /= area;
      div0  /= area;

      grf->EP_green_re( v, div, J0, div0,
                              plist[ tlist[qtri].p1 ],
                                 plist[ tlist[qtri].p2 ],
                                    plist[ tlist[qtri].p3 ], ap );
      A[0] += v[0];
      A[1] += v[1];
      A[2] += v[2];
      phi += div;
   }

   }
//  
// Iinc
   else if ( e2 >= (int)velist.size() &&
             e2 < (int)(velist.size()+Iinclist.size()) )
   {

   Iinc *ici = &Iinclist[ e2 - LTbfktlist.size() ];

   for ( int t=0; t<ici->count; ++t )
   {


   qtri = ici->trino[t];
   triangle_center( qtri, qp );

// edge 1
   if ( HYPOT( ap.x-qp.x, ap.y-qp.y, ap.z-qp.z ) < sgr )
   {
      singu = true;
      J0[0] = ( qp.x - plist[ ici->tp1[t] ].x ) * 0.5 * ici->angle[t];
      J0[1] = ( qp.y - plist[ ici->tp1[t] ].y ) * 0.5 * ici->angle[t];
      J0[2] = ( qp.z - plist[ ici->tp1[t] ].z ) * 0.5 * ici->angle[t];
      div0 = cONE * ici->angle[t];
   }
   else
      singu = false;

   for ( int ip2=0; ip2<iI.no; ip2++ )
   {
      qp.x = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].x
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].x
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].x;
      qp.y = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].y
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].y
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].y;
      qp.z = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].z
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].z
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].z;


      rv2[0] = ( qp.x - plist[ ici->tp1[t] ].x ) * 0.5 * ici->angle[t];
      rv2[1] = ( qp.y - plist[ ici->tp1[t] ].y ) * 0.5 * ici->angle[t];
      rv2[2] = ( qp.z - plist[ ici->tp1[t] ].z ) * 0.5 * ici->angle[t];

      div = cONE * ici->angle[t];

      if ( singu )
      {
         J[0] = rv2[0];
         J[1] = rv2[1];
         J[2] = rv2[2];
         grf->EP_green_ex( v, Gs, J, div, J0, div0, ap, qp );

         A[0] += iI.weights[ip2] * v[0];  // /triangle
         A[1] += iI.weights[ip2] * v[1];  // /triangle
         A[2] += iI.weights[ip2] * v[2];  // /triangle

         phi += iI.weights[ip2] * Gs;  // /triangle
      }
      else
      {
         grf->EP_green( Gd, Gs, ap, qp );

         A[0] += iI.weights[ip2] *
                 ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /triangle
         A[1] += iI.weights[ip2] *
                 ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /triangle
         A[2] += iI.weights[ip2] *
                 ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /triangle

         phi += iI.weights[ip2] * Gs * div;  // /triangle
      }
   }

   if ( singu )
   {
      area = triangle_area(qtri);
      J0[0] /= area;
      J0[1] /= area;
      J0[2] /= area;
      div0  /= area;

      grf->EP_green_re( v, div, J0, div0,
                              plist[ tlist[qtri].p1 ],
                                 plist[ tlist[qtri].p2 ],
                                    plist[ tlist[qtri].p3 ], ap );
      A[0] += v[0];
      A[1] += v[1];
      A[2] += v[2];
      phi += div;
   }

// edge 2
   if ( singu )   // same triangle
   {
      triangle_center( qtri, qp );

      J0[0] = ( qp.x - plist[ ici->tp2[t] ].x ) * 0.5 * ici->angle[t];
      J0[1] = ( qp.y - plist[ ici->tp2[t] ].y ) * 0.5 * ici->angle[t];
      J0[2] = ( qp.z - plist[ ici->tp2[t] ].z ) * 0.5 * ici->angle[t];
      div0 = cONE * ici->angle[t];
   }

   for ( int ip2=0; ip2<iI.no; ip2++ )
   {
      qp.x = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].x
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].x
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].x;
      qp.y = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].y
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].y
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].y;
      qp.z = iI.lambda1[ip2] * plist[ tlist[qtri].p1 ].z
             + iI.lambda2[ip2] * plist[ tlist[qtri].p2 ].z
             + iI.lambda3[ip2] * plist[ tlist[qtri].p3 ].z;


      rv2[0] = ( qp.x - plist[ ici->tp2[t] ].x ) * 0.5 * ici->angle[t];
      rv2[1] = ( qp.y - plist[ ici->tp2[t] ].y ) * 0.5 * ici->angle[t];
      rv2[2] = ( qp.z - plist[ ici->tp2[t] ].z ) * 0.5 * ici->angle[t];

      div = cONE * ici->angle[t];

      if ( singu )
      {
         J[0] = rv2[0];
         J[1] = rv2[1];
         J[2] = rv2[2];
         grf->EP_green_ex( v, Gs, J, div, J0, div0, ap, qp );

         A[0] += iI.weights[ip2] * v[0];  // /triangle
         A[1] += iI.weights[ip2] * v[1];  // /triangle
         A[2] += iI.weights[ip2] * v[2];  // /triangle

         phi += iI.weights[ip2] * Gs;  // /triangle
      }
      else
      {
         grf->EP_green( Gd, Gs, ap, qp );

         A[0] += iI.weights[ip2] *
                 ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /triangle
         A[1] += iI.weights[ip2] *
                 ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /triangle
         A[2] += iI.weights[ip2] *
                 ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /triangle

         phi += iI.weights[ip2] * Gs * div;  // /triangle
      }
   }

   if ( singu )
   {
      area = triangle_area(qtri);
      J0[0] /= area;
      J0[1] /= area;
      J0[2] /= area;
      div0  /= area;

      grf->EP_green_re( v, div, J0, div0,
                              plist[ tlist[qtri].p1 ],
                                 plist[ tlist[qtri].p2 ],
                                    plist[ tlist[qtri].p3 ], ap );
      A[0] += v[0];
      A[1] += v[1];
      A[2] += v[2];
      phi += div;
   }


   }  // end:t

   }  // end:elseif

}


void ModuleData::umkugel( int ltno, double &R, double CP[3] ) const
{
   double minx, maxx, miny, maxy, minz, maxz;
   const vEDGE *RWG;
   const POINT *P;

   P = &plist[ velist[ LTbfktlist[ltno].bfkt_in_LT[0] ].p1 ];
   minx = maxx = P->x;
   miny = maxy = P->y;
   minz = maxz = P->z;

   for ( unsigned int s=0; s<LTbfktlist[ltno].count; ++s )
   {
      RWG = &velist[ LTbfktlist[ltno].bfkt_in_LT[s] ];

      P = &plist[ RWG->p1 ];
      minx = ( minx < P->x ? minx : P->x );
      miny = ( miny < P->y ? miny : P->y );
      minz = ( minz < P->z ? minz : P->z );

      maxx = ( maxx > P->x ? maxx : P->x );
      maxy = ( maxy > P->y ? maxy : P->y );
      maxz = ( maxz > P->z ? maxz : P->z );

      P = &plist[ RWG->p2 ];
      minx = ( minx < P->x ? minx : P->x );
      miny = ( miny < P->y ? miny : P->y );
      minz = ( minz < P->z ? minz : P->z );

      maxx = ( maxx > P->x ? maxx : P->x );
      maxy = ( maxy > P->y ? maxy : P->y );
      maxz = ( maxz > P->z ? maxz : P->z );

      P = &plist[ RWG->spp ];
      minx = ( minx < P->x ? minx : P->x );
      miny = ( miny < P->y ? miny : P->y );
      minz = ( minz < P->z ? minz : P->z );

      maxx = ( maxx > P->x ? maxx : P->x );
      maxy = ( maxy > P->y ? maxy : P->y );
      maxz = ( maxz > P->z ? maxz : P->z );

      P = &plist[ RWG->spm ];
      minx = ( minx < P->x ? minx : P->x );
      miny = ( miny < P->y ? miny : P->y );
      minz = ( minz < P->z ? minz : P->z );

      maxx = ( maxx > P->x ? maxx : P->x );
      maxy = ( maxy > P->y ? maxy : P->y );
      maxz = ( maxz > P->z ? maxz : P->z );
   }

   CP[0] = 0.5 * ( maxx + minx );
   CP[1] = 0.5 * ( maxy + miny );
   CP[2] = 0.5 * ( maxz + minz );

   R = 0.5 * HYPOT( maxx-minx, maxy-miny, maxz-minz );
}

