//
// Class for sparse matrix
// (c) 16.3.2007 Ingo D. Rullhusen
//

#include <complex>

#include <sys/stat.h>

#include "sparse_matrix.h"
#include "sep_val.h"


template<typename data_type>
SparseMatrix<data_type>::SparseMatrix()
{
   symmU = false;
   nofrows = totlen = 0;
   first_ele = new int[1];
   first_ele[0] = 0;
   colid = NULL;
   ele = NULL;  
}


template<typename data_type>
SparseMatrix<data_type>::SparseMatrix( bool symmU_, int nofrows_, int totlen_ )
: symmU(symmU_), nofrows(nofrows_), totlen(totlen_)
{
   first_ele = new int[nofrows+1];
   for ( int i=0; i<nofrows; i++ ) first_ele[i] = -1;
   first_ele[nofrows] = totlen;   // dummy at end
   colid = new int[totlen];
   ele = new data_type[totlen];
   for ( int i=0; i<totlen; i++ )
   {
      colid[i] = -1;
      ele[i] = 0;
   }
}


template<typename data_type>
SparseMatrix<data_type>::SparseMatrix( fstream &bfile )
{
   bfile.read( (char*)&symmU, sizeof(symmU) );
   bfile.read( (char*)&nofrows, sizeof(nofrows) );
   bfile.read( (char*)&totlen, sizeof(totlen) );
   if ( bfile.bad() )
   {
      symmU = false;
      nofrows = totlen = 0;
   }

   first_ele = new int[nofrows+1];
   bfile.read( (char*) first_ele, sizeof(*first_ele)*(nofrows+1) );
   if ( bfile.bad() )
   {
      first_ele[0] = 0;
   }

   if ( totlen > 0 )
   {
      colid = new int[totlen];
      bfile.read( (char*) colid, sizeof(*colid)*totlen );

      ele = new data_type[totlen];
      bfile.read( (char*) ele, sizeof(*ele)*totlen );
   }
   else
   {
      colid = NULL;
      ele = NULL;
   }
}


template<typename data_type>
SparseMatrix<data_type>::SparseMatrix( const SparseMatrix<data_type> *SMin )
{
   symmU = SMin->symmU;
   nofrows = SMin->nofrows;
   totlen = SMin->totlen;

   first_ele = new int[nofrows+1];
   for ( int i=0; i<nofrows+1; i++ ) first_ele[i] = SMin->first_ele[i];

   colid = new int[totlen];
   ele = new data_type[totlen];
   for ( int i=0; i<totlen; i++ )
   {
      colid[i] = SMin->colid[i];
      ele[i] = SMin->ele[i];
   }
}


template<typename data_type>
SparseMatrix<data_type>::~SparseMatrix()
{
   delete [] first_ele;
   delete [] colid;
   delete [] ele;
}


template<typename data_type>
void SparseMatrix<data_type>::vecmult( data_type *Vin, data_type *Vout ) const
{
   int i;

   if ( symmU )
   {

#pragma omp parallel for
      for ( i=0; i<nofrows; i++ )
      {
         Vout[i] = ele[ first_ele[i] ] * Vin[i];
         for ( int j=first_ele[i]+1; j<first_ele[i+1]; j++ )
         {
            Vout[i] += ele[j] * Vin[ colid[j] ];
         }
      }

      for ( i=0; i<nofrows; i++ )
      {
         for ( int j=first_ele[i]+1; j<first_ele[i+1]; j++ )
         {
            Vout[ colid[j] ] += ele[j] * Vin[i];
         }
      }

   }
   else
   {
#pragma omp parallel for
      for ( i=0; i<nofrows; i++ )
      {
         Vout[i] = ele[ first_ele[i] ] * Vin[ colid[ first_ele[i] ] ];
         for ( int j=first_ele[i]+1; j<first_ele[i+1]; j++ )
         {
            Vout[i] += ele[j] * Vin[ colid[j] ];
         }
      }   
   }

}


template<typename data_type>
data_type SparseMatrix<data_type>::operator()( int index1, int index2 ) const
{
   int id1, id2;
#ifdef BOUNDARY_CHECK
   if ( index1 < 0 || index1 >= nofrows || index2 < 0 || index2 >= nofrows )
   {
      throw( "Error in SparseMatrix::operator(): index out of bounds!" );
   }
#endif
   if ( index1 > index2 && symmU )
   {
      id1 = index2;
      id2 = index1;
   }
   else
   {
      id1 = index1;
      id2 = index2;
   }
   for ( int i=first_ele[id1]; i<first_ele[id1+1]; i++ )
   {
      if ( colid[i] == id2 ) return( ele[i] );
   }
   return(0);
}


template<typename data_type>
void SparseMatrix<data_type>::to_binary( fstream &bfile ) const
{
   bfile.write( (char*)&symmU, sizeof(symmU) );
   bfile.write( (char*)&nofrows, sizeof(nofrows) );
   bfile.write( (char*)&totlen, sizeof(totlen) );

   bfile.write( (char*) first_ele, sizeof(*first_ele)*(nofrows+1) );
   bfile.write( (char*) colid, sizeof(*colid)*totlen );
   bfile.write( (char*) ele, sizeof(*ele)*totlen );
}


template<typename data_type>
void SparseMatrix<data_type>::profile_out( string filename ) const
{
   fstream dfile;

   dfile.open( filename.c_str(), ios::out );
   if ( !dfile.is_open() )
   {
      throw( "Error in SparseMatrix::profile_out(): cannot open file" );
   }

   for ( int i=0; i<nofrows; i++ )
      for ( int k=first_ele[i]; k<first_ele[i+1]; k++ )
      {
         dfile << colid[k] << " " << i << endl;
      }

   dfile.close();
}


//template class SparseMatrix< complex<float> >;
//template class SparseMatrix< complex<double> >;
template class SparseMatrix< SEP_Z_VAL >;
